-- The alphabet of a process; soundness and completeness.
module Alphabet where

open import Calculus
open import Data.List
open import Refinement hiding (preserved)
open import Theory

-- A term of type Alphabet p a is a proof that a is in the alphabet of p.
data Alphabet : Process → A → Set where
  then₁ : ∀ {p q a} →
    Alphabet p a →
    Alphabet (p # q) a

  then₂ : ∀ {p q a} →
    Alphabet q a →
    Alphabet (p # q) a

  +₁ : ∀ {p q a} →
    Alphabet p a →
    Alphabet (p + q) a

  +₂ : ∀ {p q a} →
    Alphabet q a →
    Alphabet (p + q) a

  ||₁ : ∀ {p q a} →
    Alphabet p a →
    Alphabet (p || q) a

  ||₂ : ∀ {p q a} →
    Alphabet q a →
    Alphabet (p || q) a

  star : ∀ {p a} →
    Alphabet p a →
    Alphabet (p *) a

  event : ∀ {r a} →
    Alphabet (event a r) a

-- Progress.
initial : ∀ {a p} →
  p / a →
  Alphabet p a
initial (then₁ e) = then₁ (initial e)
initial (then₂ φ e) = then₂ (initial e)
initial (||₁ e) = ||₁ (initial e)
initial (||₂ e) = ||₂ (initial e)
initial (+₁ e) = +₁ (initial e)
initial (+₂ e) = +₂ (initial e)
initial (star e) = star (initial e)
initial event = event

-- Preservation.
preserved : ∀ {a p b} →
  (e : p / a) →
  Alphabet (next e) b →
  Alphabet p b
preserved (then₁ e) (then₁ α) = then₁ (preserved e α)
preserved (then₁ e) (then₂ α) = then₂ α
preserved (then₂ φ e) α = then₂ (preserved e α)
preserved (||₁ e) (||₁ α) = ||₁ (preserved e α)
preserved (||₁ e) (||₂ α) = ||₂ α
preserved (||₂ e) (||₁ α) = ||₁ α
preserved (||₂ e) (||₂ α) = ||₂ (preserved e α)
preserved (+₁ e) α = +₁ (preserved e α)
preserved (+₂ e) α = +₂ (preserved e α)
preserved (star e) (then₁ α) = star (preserved e α)
preserved (star e) (then₂ α) = α
preserved event ()

-- example (_ : Alphabet p a) is a trace after which p can a-reduce,
-- i.e. a witness to a being in p's alphabet.
example : ∀ {p a} → Alphabet p a → List A
example (then₁ p) = example p
example {p # q} (then₂ α) = inhabitant p ++ example α
example (+₁ p) = example p
example (+₂ p) = example p
example (||₁ p) = example p
example (||₂ p) = example p
example (star p) = example p
example event = []

-- Completeness.
complete : ∀ {p a} (α : Alphabet p a) → Reaches (consumes a) p (example α)
complete (then₁ α) = #-reachable then₁ (complete α)
complete {p # q} (then₂ α) = #-sequence then₂ (witness p) (complete α)
complete (+₁ α) = inject +₁ (complete α)
complete (+₂ α) = inject +₂ (complete α)
complete (||₁ α) = inject #-|| (#-reachable then₁ (complete α))
complete (||₂ α) = inject (<=-trans #-|| ||-comm) (#-reachable then₁ (complete α))
complete (star α) = inject star (complete α)
complete event = stop event
