-- Formalisation of the process calculus and operational semantics.
module Calculus where

open import Data.Empty
open import Relation.Unary using (Decidable)
open import Relation.Nullary

infix 10 _/_
infixr 20 _#_
infixr 20 _+_
infixr 20 _||_

-- The types of calls and results.
postulate
  A R : Set

-- Processes.
data Process : Set where
  -- Empty string.
  ε : Process
  -- Sequential composition.
  _#_ : (p q : Process) → Process
  -- Parallel composition.
  _||_   : (p q : Process) → Process
  -- Choice.
  _+_    : (p q : Process) → Process
  -- Replication.
  _*     : (p   : Process) → Process
  -- A single event.
  event  : (a : A) (r : R) → Process

-- A value of type Final p is a proof that p is final.
data Final : Process → Set where
  _#_ : ∀ {p q} →
    (φ₁ : Final p) →
    (φ₂ : Final q) →
    Final (p # q)

  _||_ : ∀ {p q} →
    (φ₁ : Final p) →
    (φ₂ : Final q) →
    Final (p || q)

  +₁ : ∀ {p q} →
    (φ : Final p) →
    Final (p + q)

  +₂ : ∀ {p q} →
    (φ : Final q) →
    Final (p + q)

  star : ∀ {p} → Final (p *)

  ε : Final ε

-- A value of type p / a is a proof that p can a-reduce.
-- We define the process that p reduces to in the function 'next'.
data _/_ : Process → A → Set where
  then₁ : ∀ {a p q} →
    (e : p / a) →
    p # q / a

  then₂ : ∀ {a p q} →
    -- We use irrelevance for φ so that Agda doesn't consider it when
    -- deciding whether two reductions are equal.
    .(φ : Final p) →
    (e : q / a) →
    p # q / a

  ||₁ : ∀ {a p q} →
    (e : p / a) →
    p || q / a

  ||₂ : ∀ {a p q} →
    (e : q / a) →
    p || q / a

  +₁ : ∀ {a p q} →
    (e : p / a) →
    p + q / a

  +₂ : ∀ {a p q} →
    (e : q / a) →
    p + q / a

  star : ∀ {a p} →
    (e : p / a) →
    p * / a

  event : ∀ {a r} →
    event a r / a

Consumes : Process → A → Set
Consumes p a = p / a

-- The next process reached after a reduction.
next : ∀ {p a} → p / a → Process
next {p # q} (then₁ e) = next e # q
next (then₂ φ e) = next e
next {p || q} (||₁ e) = next e || q
next {p || q} (||₂ e) = p || next e
next (+₁ e) = next e
next (+₂ e) = next e
next {p *} (star e) = next e # p *
next event = ε

-- Compute whether a process is final or not.
-- This and all the machinery below is needed because of our
-- use of irrelevance in the definition of then₂ above.
decide-φ : Decidable Final
decide-φ ε = yes ε
decide-φ (p # q) with decide-φ p | decide-φ q
... | yes φ₁ | yes φ₂ = yes (φ₁ # φ₂)
... | _ | no ¬φ = no (λ{(φ₁ # φ₂) → ¬φ φ₂})
... | no ¬φ | _ = no (λ{(φ₁ # φ₂) → ¬φ φ₁})
decide-φ (p || q) with decide-φ p | decide-φ q
... | yes φ₁ | yes φ₂ = yes (φ₁ || φ₂)
... | no ¬φ | _ = no (λ{(φ₁ || φ₂) → ¬φ φ₁})
... | _ | no ¬φ = no (λ{(φ₁ || φ₂) → ¬φ φ₂})
decide-φ (p + q) with decide-φ p | decide-φ q
... | yes φ | _ = yes (+₁ φ)
... | _ | yes φ = yes (+₂ φ)
... | no ¬φ₁ | no ¬φ₂ = no (λ{(+₁ φ) → ¬φ₁ φ
                            ;(+₂ φ) → ¬φ₂ φ})
decide-φ (p *) = yes star
decide-φ (event a r) = no (λ ())

⊥-irrelevant-elim : {A : Set} → .⊥ → A
⊥-irrelevant-elim ()

recover : {A : Set} {F : A → Set} {x : A} → Decidable F → .(F x) → F x
recover {x = x} decide d with decide x
... | yes p = p
... | no ¬p = ⊥-irrelevant-elim (¬p d)

-- Given an irrelevant proof Final p, construct a relevant one,
-- with the help of decide-φ.
recover-φ : ∀ {p} → .(Final p) → Final p
recover-φ φ = recover decide-φ φ
