-- The completeness proof.
-- Follows the structure described in the paper:
-- first inversion lemmas, then an inductive proof.
module Completeness where

open import Calculus
open import Overlaps renaming (complete to o-complete)
open import Theory
open import Refinement
open import Typing
open import Alphabet renaming (complete to α-complete)
open import Data.List
open import Determinism

-- Inversion lemmas.
invert-then₁ : ∀ {p q} → Deterministic _//_ (p # q) → Deterministic _//_ p
invert-then₁ d es₁ es₂ =
  #-reachable-injective _ es₁ es₂ (d (#-reachable _ es₁) (#-reachable _ es₂))

invert-then₂ : ∀ {p q} → Deterministic _//_ (p # q) → Deterministic _//_ q
invert-then₂ {p = p} d es₁ es₂ =
  #-sequence-injective _ (witness p) es₁ es₂
    (d (#-sequence _ (witness p) es₁)
       (#-sequence _ (witness p) es₂))

-- For the remaining inversion lemmas we just rely on <=.
-- For example: if p || q is deterministic,
-- then p # q is deterministic (since p # q <= p || q),
-- so p is deterministic.
invert-+₁ : ∀ {p q} → Deterministic _//_ (p + q) → Deterministic _//_ p
invert-+₁ = deterministic-<= +₁

invert-+₂ : ∀ {p q} → Deterministic _//_ (p + q) → Deterministic _//_ q
invert-+₂ = deterministic-<= +₂

invert-||₁ : ∀ {p q} → Deterministic _//_ (p || q) → Deterministic _//_ p
invert-||₁ d = invert-then₁ (deterministic-<= #-|| d)

invert-||₂ : ∀ {p q} → Deterministic _//_ (p || q) → Deterministic _//_ q
invert-||₂ d = invert-then₂ (deterministic-<= #-|| d)

invert-* : ∀ {p} → Deterministic _//_ (p *) → Deterministic _//_ p
invert-* d = deterministic-<= star d

-- These lemmas represent the inductive step for each constructor.

#-ambiguity : ∀ {p q a} → (o : Overlaps p a) → q / a →
  Reaches ambiguity (p # q) (overlap o)
#-ambiguity {p} {q} {a} o e = #-reachable f (o-complete o)
  where f : ∀ {p} → Boundary a p → Ambiguity (p # q)
        f (bound φ f) = amb (then₁ f) (then₂ φ e) (λ{()})

||-ambiguity : ∀ {p q a} → (α : Alphabet p a) → (β : Alphabet q a) →
  Reaches ambiguity (p || q) (example α ++ example β)
||-ambiguity α β = ||-parallel (λ e₁ e₂ → amb (||₁ e₁) (||₂ e₂) (λ ())) (α-complete α) (α-complete β)

+-ambiguity : ∀ {p q a} → p / a → q / a →
  Reaches ambiguity (p + q) []
+-ambiguity e₁ e₂ = stop (amb (+₁ e₁) (+₂ e₂) (λ ()))

*-ambiguity : ∀ {p q a} → p / a → (o : Overlaps q a) →
  Reaches ambiguity (q # p *) (overlap o)
*-ambiguity e o = #-ambiguity o (star e)

-- Putting the completeness proof together.
complete : ∀ {p} → Deterministic _//_ p → D p
complete {p # q} d = then (complete (invert-then₁ d)) (complete (invert-then₂ d)) (λ o e → nondeterministic d (#-ambiguity o e))
complete {p || q} d = d|| (complete (invert-||₁ d)) (complete (invert-||₂ d)) (λ α β → nondeterministic d (||-ambiguity α β))
complete {p + q} d = d+ (complete (invert-+₁ d)) (complete (invert-+₂ d)) (λ e₁ e₂ → nondeterministic d (+-ambiguity e₁ e₂))
complete {p *} d = d* (complete (invert-* d)) (λ o e → nondeterministic d (star _ ∷ *-ambiguity e o) )
complete {event a r} d = event
complete {ε} d = ε
