-- A denotational semantics, and proving it's equivalent to the
-- operational semantics.
module Denotational where

open import Calculus
open import Refinement
open import Data.List
open import Relation.Binary.PropositionalEquality

-- Merge xs ys zs: zs is an interleaving of xs and ys.
data Merge {A : Set} : List A → List A → List A → Set where
  [] : Merge [] [] []
  left : ∀ {x xs ys zs} → (pf : Merge xs ys zs) → Merge (x ∷ xs) ys (x ∷ zs)
  right : ∀ {y xs ys zs} → (pf : Merge xs ys zs) → Merge xs (y ∷ ys) (y ∷ zs)

mutual
  -- A value of type Accepted p is a trace accepted by p.
  data Accepted (p : Process) : Set where
    acc : ∀ {as} → L p as → Accepted p

  -- The sequence associated with an Accepted.
  trace : ∀ {p} → Accepted p → List A
  trace (acc {as} _) = as

  -- The denotational semantics: a value of type L p as is a proof
  -- that as is in the language of p.
  data L : Process → List A → Set where
    ε : L ε []
    event : ∀ {a r} → L (event a r) (a ∷ [])
    then : ∀ {as bs p q cs} → (l₁ : L p as) → (l₂ : L q bs) → (pf : as ++ bs ≡ cs) → L (p # q) cs
    par : ∀ {as bs p q cs} → (l₁ : L p as) → (l₂ : L q bs) → (pf : Merge as bs cs) → L (p || q) cs
    +₁ : ∀ {p q as} → (l : L p as) → L (p + q) as
    +₂ : ∀ {p q as} → (l : L q as) → L (p + q) as
    star : ∀ {as p} (accs : List (Accepted p)) → (pf : concatMap trace accs ≡ as) → L (p *) as

-- The notion of finality matches up with the denotational semantics.
φ-sound : ∀ {p} → Final p → L p []
φ-sound (φ₁ # φ₂) = then (φ-sound φ₁) (φ-sound φ₂) refl
φ-sound (φ₁ || φ₂) = par (φ-sound φ₁) (φ-sound φ₂) []
φ-sound (+₁ φ) = +₁ (φ-sound φ)
φ-sound (+₂ φ) = +₂ (φ-sound φ)
φ-sound star = star [] refl
φ-sound ε = ε

φ-complete : ∀ {p} → L p [] → Final p
φ-complete ε = ε
φ-complete (then {[]} {[]} l₁ l₂ refl) = φ-complete l₁ # φ-complete l₂
φ-complete (then {[]} {_ ∷ _} l₁ l₂ ())
φ-complete (then {_ ∷ _} {[]} l₁ l₂ ())
φ-complete (then {_ ∷ _} {_ ∷ _} l₁ l₂ ())
φ-complete (par l₁ l₂ []) = φ-complete l₁ || φ-complete l₂
φ-complete (+₁ l) = +₁ (φ-complete l)
φ-complete (+₂ l) = +₂ (φ-complete l)
φ-complete (star accs pf) = star

-- One-step reduction respects the denotational semantics.
/-sound : ∀ {p a as} → (e : p / a) → L (next e) as → L p (a ∷ as)
/-sound {a = a} (then₁ e) (then l₁ l₂ pf) = then (/-sound e l₁) l₂ (cong (λ as → a ∷ as) pf)
/-sound (then₂ φ e) l = then (φ-sound (recover-φ φ)) (/-sound e l) refl
/-sound (||₁ e) (par l₁ l₂ pf) = par (/-sound e l₁) l₂ (left pf)
/-sound (||₂ e) (par l₁ l₂ pf) = par l₁ (/-sound e l₂) (right pf)
/-sound (+₁ e) l = +₁ (/-sound e l)
/-sound (+₂ e) l = +₂ (/-sound e l)
/-sound {a = a} (star e) (then {as} l₁ (star accs pf₁) pf₂) =
  star (acc (/-sound e l₁) ∷ accs)
    (cong (λ as → a ∷ as)
      (trans (cong (λ bs → as ++ bs) pf₁) pf₂))
/-sound event ε = event

/-complete-1 : ∀ {p a as} → L p (a ∷ as) → p / a
/-complete-1 event = event
/-complete-1 (then {[]} {[]} l₁ l₂ ())
/-complete-1 (then {[]} {_ ∷ _} l₁ l₂ refl) = then₂ (φ-complete l₁) (/-complete-1 l₂)
/-complete-1 (then {_ ∷ _} l₁ l₂ refl) = then₁ (/-complete-1 l₁)
/-complete-1 (par l₁ l₂ (left pf)) = ||₁ (/-complete-1 l₁)
/-complete-1 (par l₁ l₂ (right pf)) = ||₂ (/-complete-1 l₂)
/-complete-1 (+₁ l) = +₁ (/-complete-1 l)
/-complete-1 (+₂ l) = +₂ (/-complete-1 l)
/-complete-1 (star [] ())
/-complete-1 (star (acc {[]} l ∷ accs) pf) = /-complete-1 (star accs pf)
/-complete-1 (star (acc {_ ∷ _} l ∷ accs) refl) = star (/-complete-1 l)

/-complete-2 : ∀ {p a as} → (l : L p (a ∷ as)) → L (next (/-complete-1 l)) as
/-complete-2 event = ε
/-complete-2 (then {[]} {[]} l₁ l₂ ())
/-complete-2 (then {[]} {_ ∷ _} l₁ l₂ refl) = /-complete-2 l₂
/-complete-2 (then {_ ∷ _} l₁ l₂ refl) = then (/-complete-2 l₁) l₂ refl
/-complete-2 (par l₁ l₂ (left pf)) = par (/-complete-2 l₁) l₂ pf
/-complete-2 (par l₁ l₂ (right pf)) = par l₁ (/-complete-2 l₂) pf
/-complete-2 (+₁ l) = /-complete-2 l
/-complete-2 (+₂ l) = /-complete-2 l
/-complete-2 (star [] ())
/-complete-2 (star (acc {[]} l ∷ accs) pf) = /-complete-2 (star accs pf)
/-complete-2 (star (acc {_ ∷ _} l ∷ accs) refl) = then (/-complete-2 l) (star accs refl) refl

-- The operational and denotational semantics are equivalent.
∈-sound : ∀ {p as} → as ∈ p → L p as
∈-sound (stop φ) = φ-sound (recover-φ φ)
∈-sound (e ∷ es) = /-sound e (∈-sound es)

∈-complete : ∀ {p as} → L p as → as ∈ p
∈-complete {as = []} l = stop (φ-complete l)
∈-complete {as = _ ∷ _} l = /-complete-1 l ∷ ∈-complete (/-complete-2 l)
