-- Definition of determinism (semantically) and some lemmas.
module Determinism where

open import Calculus
open import Refinement renaming
  (progress to <=-progress;
   progress-injective to <=-progress-injective;
   preserved to <=-preserved)
open import Relation.Binary.PropositionalEquality
open import Data.Empty
open import Data.Unit

-- Deterministic _/_ p: for all a, there is only one reduction p / a.
Deterministic : {A : Set} → (Process → A → Set) → Process → Set
Deterministic _/_ p = ∀ {a} →
  (x : p / a) →
  (y : p / a) →
  x ≡ y

-- The rest of this file is only used in the completeness proof.

-- Relate multi-step reduction to single-step reduction.
d//-initial : ∀ {p} → Deterministic _//_ p → Deterministic _/_ p
d//-initial d e₁ e₂ = ≡-head (d (e₁ ∷ stop tt) (e₂ ∷ stop tt))

d//-preserved : ∀ {p a} → (e : p / a) →
  Deterministic _//_ p → Deterministic _//_ (next e)
d//-preserved e d es₁ es₂ = ≡-tail (d (e ∷ es₁) (e ∷ es₂))

-- Ambiguity p is a proof that p has two distinct reductions for the same event.
-- Thus Reaches ambiguity p means that p is nondeterministic.
--
-- We can use all the lemmas about the Reaches relation to
-- reason about ambiguity, which is why it's useful as a concept
-- in its own right.
data Ambiguity : Process → Set where
  amb : ∀ {a p} → (e₁ e₂ : p / a) → (e₁ ≡ e₂ → ⊥) → Ambiguity p

ambiguity = predicate Ambiguity respects
  where
    respects : Respects Ambiguity
    respects pf₁ (amb e₁ e₂ ≠) =
      amb (<=-progress pf₁ e₁) (<=-progress pf₁ e₂)
        (λ pf₂ → ≠ (<=-progress-injective pf₁ e₁ e₂ pf₂))

-- A process cannot be both deterministic and ambiguous.
nondeterministic : ∀ {p as} → Deterministic _//_ p → Reaches ambiguity p as → ⊥
nondeterministic {p} d (stop ambig) = ⊥-irrelevant-elim (contra ambig)
  where contra : Ambiguity p → ⊥
        contra (amb e₁ e₂ ≠) = ≠ (≡-head (d (e₁ ∷ stop tt) (e₂ ∷ stop tt)))
nondeterministic d (e ∷ es) =
  nondeterministic (d//-preserved e d) es

-- Determinism respects <=: if p <= q and q is deterministic then p is
-- deterministic.
deterministic-<= : ∀ {p q} →
  p <= q → Deterministic _//_ q → Deterministic _//_ p
deterministic-<= pf d (stop φ) (stop φ₁) = refl
deterministic-<= pf d (e₁ ∷ es₁) (e₂ ∷ es₂) with <=-progress-injective pf e₁ e₂ (d//-initial d (<=-progress pf e₁) (<=-progress pf e₂))
deterministic-<= pf d (e ∷ es₁) (.e ∷ es₂) | refl =
  cong (λ es → e ∷ es)
    (deterministic-<= (<=-preserved pf e) (d//-preserved (<=-progress pf e) d) es₁ es₂)
