-- The "?" overlaps judgement; soundness and completeness.
module Overlaps where

open import Calculus
open import Data.List
open import Theory
open import Refinement hiding (preserved)

-- A term of type Overlaps p a is a proof that a overlaps p.
data Overlaps : Process → A → Set where
  then₁ : ∀ {a p q} →
    (φ : Final q) →
    (o : Overlaps p a) →
    Overlaps (p # q) a

  then₂ : ∀ {a p q} →
    (o : Overlaps q a) →
    Overlaps (p # q) a

  +₁ : ∀ {a p q} →
    (o : Overlaps p a) →
    Overlaps (p + q) a

  +₂ : ∀ {a p q} →
    (o : Overlaps q a) →
    Overlaps (p + q) a

  +φ₁ : ∀ {a p q} →
    (e : p / a) →
    (φ : Final q) →
    Overlaps (p + q) a

  +φ₂ : ∀ {a p q} →
    (e : q / a) →
    (φ : Final p) →
    Overlaps (p + q) a

  ||₁ : ∀ {a p q} →
    (o : Overlaps p a) →
    Overlaps (p || q) a

  ||₂ : ∀ {a p q} →
    (o : Overlaps q a) →
    Overlaps (p || q) a

  *₁ : ∀ {a p} →
    (o : Overlaps p a) →
    Overlaps (p *) a

  *₂ : ∀ {a p} →
    (e : p / a) →
    Overlaps (p *) a

-- Progress.
initial : ∀ {a p} →
  p / a →
  Final p →
  Overlaps p a
initial (then₁ e) (φ₁ # φ₂) = then₁ φ₂ (initial e φ₁)
initial (then₂ φ e) (φ₁ # φ₂) = then₂ (initial e φ₂)
initial (||₁ e) (φ₁ || φ₂) = ||₁ (initial e φ₁)
initial (||₂ e) (φ₁ || φ₂) = ||₂ (initial e φ₂)
initial (+₁ e) (+₁ φ) = +₁ (initial e φ)
initial (+₁ e) (+₂ φ) = +φ₁ e φ
initial (+₂ e) (+₁ φ) = +φ₂ e φ
initial (+₂ e) (+₂ φ) = +₂ (initial e φ)
initial (star e) star = *₂ e
initial event ()

-- Preservation.
preserved : ∀ {a p b} → (e : p / a) → Overlaps (next e) b → Overlaps p b
preserved (then₁ e) (then₁ φ o) = then₁ φ (preserved e o)
preserved (then₁ e) (then₂ o) = then₂ o
preserved (then₂ φ e) o = then₂ (preserved e o)
preserved (||₁ e) (||₁ o) = ||₁ (preserved e o)
preserved (||₁ e) (||₂ o) = ||₂ o
preserved (||₂ e) (||₁ o) = ||₁ o
preserved (||₂ e) (||₂ o) = ||₂ (preserved e o)
preserved (+₁ e) o = +₁ (preserved e o)
preserved (+₂ e) o = +₂ (preserved e o)
preserved (star e) (then₁ φ o) = *₁ (preserved e o)
preserved (star e) (then₂ o) = o
preserved event ()

-- Boundary a p means that p is final and a-reduces.
-- Thus Reaches (boundary a) p means that a semantically overlaps with p.
data Boundary : A → Process → Set where
  bound : ∀ {a p} → .(φ : Final p) → (e : p / a) → Boundary a p

boundary : A → Predicate
boundary a = predicate (Boundary a) respects
  where
    respects : Respects (Boundary a)
    respects pf (bound φ e) =
      bound (respects-<= final pf φ) (respects-<= (consumes _) pf e)

-- Given that a overlaps p, construct a trace that witnesses it.
overlap : ∀ {p a} → Overlaps p a → List A
overlap (then₁ φ o) = overlap o
overlap {p # q} (then₂ o) = inhabitant p ++ overlap o
overlap (+₁ o) = overlap o
overlap (+₂ o) = overlap o
overlap (+φ₁ f φ) = []
overlap (+φ₂ f φ) = []
overlap {p || q} (||₁ o) = inhabitant q ++ overlap o
overlap {p || q} (||₂ o) = inhabitant p ++ overlap o
overlap (*₁ o) = overlap o
overlap {p *}(*₂ f) = []

-- If a semantically overlaps q then it semantically overlaps p # q.
#-boundary-reachable : ∀ {a as p q} →
  Reaches (boundary a) q as → Reaches (boundary a) (p # q) (inhabitant p ++ as)
#-boundary-reachable {a} {p = p} es = #-sequence (λ {p}{q} → f {p}{q}) (witness p) es
  where f : ∀ {p q} → .(Final p) → Boundary a q → Boundary a (p # q)
        f φ₁ (bound φ₂ e)  = bound (φ₁ # φ₂) (then₂ φ₁ e)

-- Completeness of the overlaps judgement.
complete : ∀ {p a} → (o : Overlaps p a) → Reaches (boundary a) p (overlap o)
complete (then₁ φ o) = inject (#₁ φ) (complete o)
complete {p # q} (then₂ o) = #-boundary-reachable (complete o)
complete (+₁ o) = inject +₁ (complete o)
complete (+₂ o) = inject +₂ (complete o)
complete (+φ₁ f φ) = stop (bound (+₂ φ) (+₁ f))
complete (+φ₂ f φ) = stop (bound (+₁ φ) (+₂ f))
complete {p || q} (||₁ o) = inject (<=-trans #-|| ||-comm) (#-boundary-reachable (complete o))
complete {p || q} (||₂ o) = inject #-|| (#-boundary-reachable (complete o))
complete (*₁ o) = inject star (complete o)
complete {p *} {a} (*₂ f) = stop (bound star (star f))
