-- In this module we define a type Reaches P p as.
--
-- Reaches P p as, where P is a predicate on processes, p is a process,
-- and 'as' is a trace, means that 'as' can reduce p to a process that
-- satisfies P.
--
-- Why is this useful? We can represent many semantic properties of
-- processes by Reaches:
--   * If P is the predicate that's true everywhere,
--     then Reaches P p as means that p can consume the events
--     'as' without failing.
--   * If P is the predicate that's true of final processes,
--     then Reaches P p as means that p accepts as.
--   * If P a is a predicate that's true when a process a-reduces,
--     then Reaches (P a) p as means that a is semantically in the
--     alphabet of p, witnessed by the trace as ++ (a ∷ [])
--   * If P a is a predicate that's true when a process has two distinct
--     a-reductions, then Reaches (P a) p as means that p is ambiguous
--     when given the trace as ++ (a ∷ [])
--   * If P a is a predicate that's true when a process is final and a-reduces,
--     then Reaches (P a) p as witnesses that a semantically overlaps p.
--
-- We also define a refinement relation p <= q on processes, to mean that the
-- traces of p (including incomplete ones) are a subset of the traces of q.
-- For example, p <= p + q because anything that p accepts p+q also accepts;
-- p # q <= p || q because sequential composition is a more restrictive case
-- of parallel composition; etc.
--
-- The main result of this module is that Reaches P respects <=:
--   If p <= q and Reaches P p then Reaches P q.
-- This is true provided that P itself respects <=. We prove in this file that
-- finality and reduction respect <=; our predicates in the rest of the
-- development are combinations of those two, so then it's trivial
-- to prove that they too respect <=.
--
-- This result is very useful. For example, consider the inversion
-- lemmas for the completeness proof: if p or q is ambiguous then
-- p # q, p || q, p + q, p * are ambiguous.
-- We observe that p <= p + q and q <= p + q; thus if p or q is ambiguous
-- then so is p + q.
-- Also, p <= p *; thus if p is ambiguous then so is p *.
-- We must prove that if p or q is ambiguous then so is p # q.
-- But once we've done that we can observe that since p # q <= p || q,
-- p || q must also be ambiguous.
-- Thus we get to skip an awful lot of the semantic reasoning in the
-- completeness proof.
module Refinement where

open import Calculus
open import Relation.Binary.PropositionalEquality
open import Data.List
open import Data.Unit

infix 2 _<=_

-- The <= relation on processes.
data _<=_ : Process → Process → Set where
  +₁ : ∀ {p q} → p <= p + q
  +₂ : ∀ {p q} → q <= p + q
  star : ∀ {p} → p <= p *
  #-|| : ∀ {p q} → p # q <= p || q
  ||-comm : ∀ {p q} → p || q <= q || p
  #₁ : ∀ {p q} → (φ : Final q) → p <= p # q
  #₂ : ∀ {p q} → (φ : Final p) → q <= p # q
  <=-refl : ∀ {p} → p <= p
  <=-trans : ∀ {p q r} → p <= q → q <= r → p <= r

infix 2 _//_
infix 2 _∈_

-- The predicate P must itself respect <=.
Respects : (Process → Set) → Set
Respects P = ∀ {p q} → p <= q → P p → P q

-- A predicate together with a proof that it respects <=.
record Predicate : Set₁ where
  constructor predicate
  field
    eval : Process → Set
    respects-<= : Respects eval
open Predicate public

-- The Reaches type.
data Reaches (P : Predicate) : Process → List A → Set where
  stop : ∀ {p} → .(φ : eval P p) → Reaches P p []
  _∷_ : ∀ {a as p} →
    (e : p / a) →
    Reaches P (next e) as →
    Reaches P p (a ∷ as)

-- The Final predicate respects <=.
final = predicate Final respects
  where
    respects : Respects Final
    respects +₁ φ = +₁ φ
    respects +₂ φ = +₂ φ
    respects star φ = star
    respects #-|| (φ₁ # φ₂) = φ₁ || φ₂
    respects ||-comm (φ₁ || φ₂) = φ₂ || φ₁
    respects (#₁ φ₁) φ₂ = φ₂ # φ₁
    respects (#₂ φ₁) φ₂ = φ₁ # φ₂
    respects <=-refl φ = φ
    respects (<=-trans pf₁ pf₂) φ = respects pf₂ (respects pf₁ φ)

Member = Reaches final

-- as ∈ p iff Reaches final p as.
_∈_ : List A → Process → Set
as ∈ p = Member p as

-- p / a respects <=.
consumes : ∀ a → Predicate
consumes a = predicate (λ p → p / a) respects
  where
    respects : Respects (λ p → p / a)
    respects +₁ e = +₁ e
    respects +₂ e = +₂ e
    respects star e = star e
    respects #-|| (then₁ e) = ||₁ e
    respects #-|| (then₂ φ e) = ||₂ e
    respects ||-comm (||₁ e) = ||₂ e
    respects ||-comm (||₂ e) = ||₁ e
    respects (#₁ φ) e = then₁ e
    respects (#₂ φ) e = then₂ φ e
    respects <=-refl e = e
    respects (<=-trans pf₁ pf₂) e = respects pf₂ (respects pf₁ e)

_//_ : Process → List A → Set
_//_ = Reaches (predicate _ (λ _ _ → tt))

-- Some icky definitions to help Agda's equality reasoning.
Cons : ∀ {P p a as} →
  (e : p / a) →
  (es : Reaches P (next e) as) →
  Reaches P p (a ∷ as)
Cons = _∷_

≡-head : ∀ {P p a e₁ e₂ as es₁ es₂} → (Cons{P}{p}{a}{as} e₁ es₁ ≡ e₂ ∷ es₂) → e₁ ≡ e₂
≡-head refl = refl

≡-tail : ∀ {P p a e as es₁ es₂} → (Cons{P}{p}{a}{as} e es₁ ≡ e ∷ es₂) → es₁ ≡ es₂
≡-tail refl = refl

-- A progress-like lemma for <=.
progress : ∀ {p q a} → p <= q → p / a → q / a
progress {a = a} pf e = respects-<= (consumes a) pf e

-- Essentially: <= preserves determinism.
progress-injective : ∀ {p q a} → (pf : p <= q) →
  (e₁ e₂ : p / a) →
  progress pf e₁ ≡ progress pf e₂ →
  e₁ ≡ e₂
progress-injective +₁ e .e refl = refl
progress-injective +₂ e .e refl = refl
progress-injective star e .e refl = refl
progress-injective #-|| (then₁ e) (then₁ .e) refl = refl
progress-injective #-|| (then₁ e₁) (then₂ φ e₂) ()
progress-injective #-|| (then₂ φ e₁) (then₁ e₂) ()
progress-injective #-|| (then₂ _ e) (then₂ _ .e) refl = refl
progress-injective ||-comm (||₁ e) (||₁ .e) refl = refl
progress-injective ||-comm (||₁ e₁) (||₂ e₂) ()
progress-injective ||-comm (||₂ e₁) (||₁ e₂) ()
progress-injective ||-comm (||₂ e) (||₂ .e) refl = refl
progress-injective (#₁ φ) e .e refl = refl
progress-injective (#₂ φ) e .e refl = refl
progress-injective <=-refl e₁ e₂ pf = pf
progress-injective (<=-trans pf₁ pf₂) e₁ e₂ pf₃ =
  progress-injective pf₁ e₁ e₂ (progress-injective pf₂ (progress pf₁ e₁) (progress pf₁ e₂) pf₃)

-- A preservation-like lemma for <=.
preserved : ∀ {p q a} → (pf : p <= q) → (e : p / a) → next e <= next (progress pf e)
preserved +₁ e = <=-refl
preserved +₂ e = <=-refl
preserved star e = #₁ star
preserved #-|| (then₁ e) = #-||
preserved #-|| (then₂ φ e) = <=-trans (#₂ (recover-φ φ)) #-||
preserved ||-comm (||₁ e) = ||-comm
preserved ||-comm (||₂ e) = ||-comm
preserved (#₁ φ) e = #₁ φ
preserved (#₂ φ) e = <=-refl
preserved <=-refl e = <=-refl
preserved (<=-trans pf₁ pf₂) e =
  <=-trans (preserved pf₁ e) (preserved pf₂ (progress pf₁ e))

-- The main result of this file.
inject : ∀ {P p q as} → p <= q → Reaches P p as → Reaches P q as
inject {P} pf (stop φ) = stop (respects-<= P pf φ)
inject pf (e ∷ es) = progress pf e ∷ inject (preserved pf e) es

-- The 'inject' function preserves determinism.
inject-injective : ∀ {P p q as} →
  (pf : p <= q) → (es₁ es₂ : Reaches P p as) →
  inject pf es₁ ≡ inject pf es₂ →
  es₁ ≡ es₂
inject-injective pf₁ (stop φ₁) (stop φ₂) pf₂ = refl
inject-injective pf₁ (e₁ ∷ es₁) (e₂ ∷ es₂) pf₂ with progress-injective pf₁ e₁ e₂ (≡-head pf₂)
inject-injective pf₁ (e ∷ es₁) (.e ∷ es₂) pf₂ | refl =
  cong (λ es → e ∷ es)
    (inject-injective (preserved pf₁ e) es₁ es₂ (≡-tail pf₂))
