-- Miscellaneous semantic theorems about the process calculus.
module Theory where

open import Calculus
open import Refinement
open import Data.Unit
open import Data.List
open import Relation.Binary.PropositionalEquality
open import Relation.Nullary
open Predicate

module #-Sequence{P : Predicate}
  (f : ∀ {p q} → .(Final p) → eval P q → eval P (p # q)) where

  -- Roughly: if p is inhabited and q satisfies some property
  -- then p # q satisfies that property.
  -- Useful for proving properties of #.
  #-sequence : ∀ {p q as bs} →
    as ∈ p → Reaches P q bs → Reaches P (p # q) (as ++ bs)
  #-sequence (stop φ₁) (stop φ₂) = stop (f φ₁ φ₂)
  #-sequence (stop φ) (e ∷ es) = then₂ φ e ∷ ‌es
  #-sequence (e ∷ es₁) es₂ = then₁ e ∷ #-sequence es₁ es₂

  -- Use of #-sequence respects determinism.
  #-sequence-injective : ∀ {p q as bs} →
    (es₁ : as ∈ p) (es₂ es₃ : Reaches P q bs) →
    #-sequence es₁ es₂ ≡ #-sequence es₁ es₃ →
    es₂ ≡ es₃
  #-sequence-injective (stop φ₁) (stop φ₂) (stop φ₃) pf = refl
  #-sequence-injective (stop φ) (.e₁ ∷ .es₃) (e₁ ∷ es₃) refl = refl
  #-sequence-injective (e ∷ es₁) es₂ es₃ pf = #-sequence-injective es₁ es₂ es₃ (≡-tail pf)
open #-Sequence public

-- If p satisfies some property then so does p # q.
#-reachable : ∀ {P Q} {p q as}
  (f : ∀ {p} → eval P p → eval Q (p # q)) →
  Reaches P p as → Reaches Q (p # q) as
#-reachable f (stop φ) = stop (f φ)
#-reachable f (e ∷ es) = then₁ e ∷ #-reachable f es

#-reachable-injective : ∀ {P Q} {p q as}
  (f : ∀ {p} → eval P p → eval Q (p # q)) →
  (es₁ es₂ : Reaches P p as) →
  #-reachable {Q = Q} {q = q} f es₁ ≡ #-reachable f es₂ →
  es₁ ≡ es₂
#-reachable-injective f (stop φ) (stop φ₁) pf = refl
#-reachable-injective f (e ∷ es₁) (e₁ ∷ es₂) pf with ≡-head pf
#-reachable-injective f (e ∷ es₁) (.e ∷ es₂) pf | refl =
  cong (λ es → e ∷ es) (#-reachable-injective f es₁ es₂ (≡-tail pf))

module ||-Parallel{P Q R : Predicate}
  (f : ∀ {p q} → eval P p → eval Q q → eval R (p || q)) where
  -- Roughly: if p and q both satisfy a property then so does p || q.
  ||-parallel : ∀ {p q as bs} →
    Reaches P p as → Reaches Q q bs → Reaches R (p || q) (as ++ bs)
  ||-parallel (stop φ₁) (stop φ₂) = stop (f φ₁ φ₂)
  ||-parallel (stop φ) (e ∷ es) = ||₂ e ∷ ||-parallel (stop φ) es
  ||-parallel (e ∷ es₁) es₂ = ||₁ e ∷ (||-parallel es₁ es₂)
open ||-Parallel public

-- Every process is inhabited. Here is a trace that the process accepts...
inhabitant : Process → List A
inhabitant ε = []
inhabitant (p # q) = inhabitant p ++ inhabitant q
inhabitant (p || q) = inhabitant p ++ inhabitant q
inhabitant (p + q) = inhabitant p
inhabitant (p *) = []
inhabitant (event a r) = a ∷ []

-- ...and here is a proof that the process accepts the trace.
witness : ∀ p → inhabitant p ∈ p
witness ε = stop ε
witness (p # q) = #-sequence _#_ (witness p) (witness q)
witness (p || q) = ||-parallel _||_ (witness p) (witness q)
witness (p + q) = inject +₁ (witness p)
witness (p *) = stop star
witness (event a r) = event ∷ stop ε
